from fractions import Fraction
from json import JSONEncoder
from numbers import Rational

from mpmath import mpf, mpc, matrix
from sympy import Matrix

from qspd.util import InputFunction, Parity, ImaginaryRational


class DataEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, InputFunction):
            return {"parity": o.parity, "coeffs": {k: o[k] for k in o}}
        if isinstance(o, Parity):
            if o == Parity.EVEN:
                return "even"
            else:
                assert o == Parity.ODD
                return "odd"
        if isinstance(o, (mpf, mpc)):
            return repr(o)
        if isinstance(o, Fraction):
            return {"numerator": o.numerator, "denominator": o.denominator, "imaginary": False}
        if isinstance(o, ImaginaryRational):
            return {"numerator": o.numerator, "denominator": o.denominator, "imaginary": True}
        if isinstance(o, Matrix):
            return matrix(o)
        if isinstance(o, matrix):
            return o.tolist()
        return JSONEncoder.default(self, o)


def decode_data(data_list):
    def decode_record(record):
        record["tau"] = decode_float(record["tau"]) if "tau" in record else None
        record["epsilon"] = decode_float(record["epsilon"]) if "epsilon" in record else None
        record["bigA"] = decode_input_function(record["bigA"]) if "bigA" in record else None
        record["bigB"] = decode_input_function(record["bigB"]) if "bigB" in record else None
        record["bigN"] = int(record["bigN"]) if "bigN" in record else None
        record["dps"] = int(record["dps"]) if "dps" in record else None
        record["a"] = decode_rational_function(record["a"]) if "a" in record else None
        record["b"] = decode_rational_function(record["b"]) if "b" in record else None
        record["n"] = decode_n(record)
        record["roots"] = [decode_float(v) for v in record["roots"]] if "roots" in record else None
        record["bigR1"] = decode_float_list(record["bigR1"]) if "bigR1" in record else None
        record["bigR2"] = decode_float_list(record["bigR2"]) if "bigR2" in record else None
        record["bigF"] = [Matrix(2, 2, list(decode_matrix(v))) for v in
                          record["bigF"]] if "bigF" in record else None
        record["bigCs"] = {k: decode_matrix(v) for k, v in
                           record["bigCs"].items()} if "bigCs" in record else None
        record["angles"] = decode_angle_list(record["angles"]) if "angles" in record else None
        if "epsilonError" in record:
            # for the purpose of reading our old JSONs, (The error would be ~5 bc its in terms of epsilons)
            record["error"] = decode_float(record["epsilonError"]) / record["epsilon"]
        elif "error" in record:
            # New data JSONs should have error only (not in terms of epsilons)
            record["error"] = decode_float(record["error"])
        else:
            record["error"] = None
        if "times" in record and isinstance(record["times"], float):
            record["times"] = {"total": record["times"]}
        return record

    def decode_float(value):
        if isinstance(value, str):
            try:
                return mpf(value)
            except ValueError:
                return eval(value)
        return value

    def decode_input_function(obj):
        parity = decode_parity(obj["parity"])
        coeffs = {int(k): decode_float(v) for k, v in obj["coeffs"].items()}
        return InputFunction(coeffs, parity)

    def decode_parity(obj):
        if obj == "even":
            return Parity.EVEN
        else:
            assert obj == "odd"
            return Parity.ODD

    def decode_rational_function(obj):
        return {k: decode_rational(v) for k, v in obj.items()}

    def decode_rational(obj):
        if isinstance(obj, (Rational, ImaginaryRational)):
            return obj
        numerator = int(obj["numerator"])
        denominator = int(obj["denominator"])
        imaginary = bool(obj["imaginary"])
        if imaginary:
            return ImaginaryRational(numerator, denominator)
        else:
            return Fraction(numerator, denominator)

    def decode_float_list(obj):
        return [decode_float(v) for v in obj]

    def decode_matrix(obj):
        floats = [[decode_float(value) for value in row] for row in obj]
        return matrix(floats)

    def decode_angle_list(obj):
        if isinstance(obj, dict):
            return [decode_float(obj[str(i)]) for i in range(len(obj))]
        elif isinstance(obj, list):
            return [decode_float(angle) for angle in obj]

    def decode_n(record):
        if "n" in record:
            return int(record["n"])
        if "angles" in record:
            return len(record["angles"]) // 2
        return None

    return [decode_record(record) for record in data_list]



from mpmath import ceil, ln, mp, ldexp, besselj, fmul, mpc, fneg, e, workdps

from qspd.steps import step1, step2, step3, step4, step5
from qspd.util import InputFunction, Parity

name = "qspd"


def qspd(big_a, big_b, epsilon=1e-4, forced_n=None, return_matrices=False):
    big_n = max(big_a.degree, big_b.degree) if forced_n is None else forced_n
    mp.dps = int(ceil((big_n / 3) * ln(big_n / epsilon)))
    a, b, n = step1(big_a, big_b, epsilon, forced_n)
    if forced_n is not None:
        n = forced_n
    p, roots = step2(a, b, n)
    big_f = step3(a, b, n, p, roots)
    c_matrices = step4(big_f, n)
    return step5(n, c_matrices, return_matrices)


def jacobi_anger(tau, epsilon, forced_big_n=None):
    if forced_big_n is None:
        big_n = int(ceil(ldexp(e, -1) * tau - ln(epsilon)))  # equation 38
    else:
        big_n = forced_big_n
    with workdps(int(ceil((big_n / 3) * ln(big_n / epsilon)))):
        a_coefficients = {}
        b_coefficients = {}
        for k in range(0, big_n + 1, 2):
            coefficient = besselj(k, tau)
            a_coefficients[k] = coefficient
            a_coefficients[-k] = coefficient
        for k in range(1, big_n + 1, 2):
            coefficient = fmul(besselj(k, tau), mpc(0, -1), exact=True)
            b_coefficients[k] = coefficient
            b_coefficients[-k] = fneg(coefficient, exact=True)
        big_a = InputFunction(a_coefficients, Parity.EVEN, big_n)
        big_b = InputFunction(b_coefficients, Parity.ODD, big_n)
        return big_a, big_b



from collections import defaultdict
from fractions import Fraction

from mpmath import fdiv, fmul, fsub, floor, frac, fneg, almosteq, mp, ldexp, fadd, fraction, \
    polyroots, sqrt, fabs, mpc, expjpi, mpf, matrix, zeros, chop, eig, eye, polar, pi
from sympy import symbols, Matrix, I, fft

from qspd.util import ImaginaryRational, Parity, polyval_dict


def step1(big_a, big_b, epsilon, forced_n=None):
    if forced_n is None:
        big_n = max(big_a.degree, big_b.degree)
    else:
        big_n = forced_n

    # epsilon divided by n, used for precision
    epsilon_over_big_n = fdiv(epsilon, big_n, rounding='d')
    one_minus_five_epsilon = fsub(1, fmul(5, epsilon, exact=True), exact=True)

    # separates a floating-point number into integer and fractional parts
    def extract(x):
        return int(floor(x)), frac(x)

    # finds the "best" rational number between the two given floating-point numbers
    # based on https://en.wikipedia.org/wiki/Continued_fraction#Best_rational_within_an_interval
    def rational_within_range(x, y):
        x_int, x_frac = extract(x)
        y_int, y_frac = extract(y)
        if x_int == y_int:
            if x_frac == 0 or y_frac == 0:
                return x_int
            assert x_frac != y_frac
            if x_frac < y_frac:
                x_frac_inv = fdiv(1, x_frac, rounding='d')
                y_frac_inv = fdiv(1, y_frac, rounding='u')
            else:
                x_frac_inv = fdiv(1, x_frac, rounding='u')
                y_frac_inv = fdiv(1, y_frac, rounding='d')
            frac_rat = Fraction(1, rational_within_range(x_frac_inv, y_frac_inv))
            return x_int + frac_rat
        else:
            return min(x_int, y_int) + 1

    # converts a floating-point number to a rational number
    def float_to_rational(x):
        if x < 0:
            return -float_to_rational(fneg(x, exact=True))
        if almosteq(x, 0, abs_eps=epsilon_over_big_n):
            if forced_n is None or x == 0:
                return 0
        min_result = fsub(x, epsilon_over_big_n, rounding='u')
        if min_result <= 0:
            min_result = ldexp(1, -mp.dps)
        max_result = fadd(x, epsilon_over_big_n, rounding='d')
        return rational_within_range(min_result, max_result)

    # converts a Fourier coefficient from the original floating-point number to a rational one
    def convert_coefficient(x):
        complex_result = fmul(one_minus_five_epsilon, x, exact=True)
        real, imag = complex_result.real, complex_result.imag
        if imag == 0:
            return float_to_rational(real)
        elif real == 0:
            return ImaginaryRational(float_to_rational(imag))
        raise ValueError

    # converts each coefficient in the function, also returning the new number of coefficients
    def convert_function(f):
        parity = f.parity
        coefficients = defaultdict(lambda: 0)
        new_n = 0
        for i in range(0, big_n + 1):
            converted = convert_coefficient(f[i])
            if converted == 0:
                continue
            new_n = i
            coefficients[i] = converted
            if parity == Parity.EVEN:
                coefficients[-i] = converted
            else:
                coefficients[-i] = -converted
        return coefficients, new_n

    a, n_a = convert_function(big_a)
    b, n_b = convert_function(big_b)
    return a, b, max(n_a, n_b)


def step2(a, b, n):
    """
    :param a: (dicts of rational) the coeffs of a(z)
    :param b: (dicts of rational) the coeffs of b(z)
    :param n: (int) the max degree of a or b (computed in step 1)
    :param epsilon: related to the precision of the computations
    :return: p (dict of rationals): the coefficients of p(z) =  = 1 - a(z)^2 - b(z)^2
             roots (list of floats): the roots of p(z)
    """

    # Step 1: Compute the coeffs of a(t)^2 and b(t)^2
    def square_poly(f):
        """
        :param f: (defaultdict) the coeffs of polynomial f(t)
        :return: the coeffs of f(t)^2
        """
        result = {}
        for i in range(-2 * n, 2 * n + 1):  # range = [-2n, 2n]
            lo = max(i - n, -2 * n)  # lo and hi give the range of coeffs to sum over
            hi = min(i + n, 2 * n)
            result[i] = sum(f[j] * f[i - j] for j in range(lo, hi + 1))
        return result

    a2, b2 = square_poly(a), square_poly(b)  # a2 and b2 give the coeffs of a(t)^2 and b(t)^2

    # Step 2: Compute the coeffs of p = 1 - a^2 - b^2
    p = {i: -(a2[i] + b2[i]) for i in range(-2 * n, 2 * n + 1)}  # the coeffs of polynomial p
    p[0] += 1

    def to_mpf(x):
        return fraction(x.numerator, x.denominator)

    # Step 3: Find the roots of p
    max_degree = max(k for k in p.keys() if p[k] != 0)  # max degree of any term in p
    min_degree = min(k for k in p.keys() if p[k] != 0)  # min degree of any term in p
    max_steps = n ** 3 + n ** 2 * mp.prec + 50  # a bit overkill
    if all(v == 0 or k % 2 == 0 for k, v in p.items()):
        coeff_list = [to_mpf(p[i]) for i in range(max_degree, min_degree - 1, -2)]
        squared_roots = polyroots(coeff_list, maxsteps=max_steps)
        roots = []
        for r in squared_roots:
            sqrt_r = sqrt(r)
            roots.append(sqrt_r)
            roots.append(-sqrt_r)
    else:
        # Note: the highest-degree terms come first in p (ex: p[0] corresponds to the 2n-degree term)
        # the coeffs of polynomial p, stored in a list and expressed as floats (mpf)
        coeff_list = [to_mpf(p[i]) for i in range(max_degree, min_degree - 1, -1)]
        roots = polyroots(coeff_list, maxsteps=max_steps)
    return p, roots


#  Evaluating complementary polynomials c(z) and d(z) at T = {exp(2 pi i k / D) | k = 1,...,D}
#  D is a power of 2 that is larger than 2n+1
#  n is degree of input polynomials that have been reduced from coefficient approximation
#  roots is a list of roots of the polynomial 1 - a(z)^2 - b(z)^2
#  poly_dict is a dictionary of coefficients of the polynomial 1 - a(z)^2 - b(z)^2
#  a_dict and b_dict are polynomials a(z) and b(z) represented as dicts of coefficients
#  Can set ret_alpha to True if wanted to return both F and alpha, alpha is used for testing the midpoint of step 3
#  RETURNS F evaluated at points of T described above, as 4 lists
def step3(a, b, n, p, roots, big_r_1=None, big_r_2=None):
    big_c = [r for r in roots if r.imag > 0 and abs(r) < 1]
    if big_r_1 is None or big_r_2 is None:
        big_r = [r.real for r in roots if r.imag == 0 and abs(r) < 1]
        if big_r_1 is not None:
            big_r_2 = list(big_r)
            for r in big_r_1:
                big_r_2.remove(r)
        elif big_r_2 is not None:
            big_r_1 = list(big_r)
            for r in big_r_2:
                big_r_1.remove(r)
        else:
            big_r.sort(reverse=True)
            len_big_r_1 = len(big_r) // 2
            big_r_1 = big_r[:len_big_r_1]
            big_r_2 = big_r[len_big_r_1:]

    # Step 1: Build e(z) and e(1/z) symbolically, as described in eq. 13
    z = symbols("z")
    e_z = 1  # starts as multiplication identity
    for r in big_c:
        e_z *= z - 2 * r.real + (fabs(r) ** 2) / z
    for r in big_r_1:
        e_z *= z - r
    for r in big_r_2:
        e_z *= 1 / z - r

    e_1_z = e_z.subs(z, (1 / z))  # e(1/z)

    alpha_num = polyval_dict(p, mpc(1))  # evaluating at z = 1
    alpha_denom = e_z * e_1_z  # e(z)*e(1/z)  # symbolic expression
    alpha_denom = alpha_denom.evalf(subs={z: 1})  # evaluate at z = 1
    alpha = mpc(alpha_num / alpha_denom)  # a real alpha, We think alpha is constant for all z

    # Step 2: Compute F(t) for various inputs of t on the unit circle
    F = []  # stores the value of F(z) for various inputs z

    # Pauli Matrices (used below to compute F)
    Id = Matrix([[1, 0], [0, 1]])  # 2 x 2 identity matrix
    X = Matrix([[0, 1], [1, 0]])
    Y = Matrix([[0, -1 * I], [I, 0]])
    Z = Matrix([[1, 0], [0, -1]])

    # evaluates e(z), a(z), b(z), c(z), d(z), and F(z) over a set of d inputs, called x.
    d = 2 ** ((2 * n + 1).bit_length())  # next power of 2 greater than 2n + 1
    for k in range(d):  # [0, D - 1]
        x = expjpi(2 * mpf(k) / mpf(d))  # the input value (from eq. 15)

        # Compute the values of e(x), e_inv(x), a(x), b(x), c(x), d(x), and F(x)
        e_x = e_z.evalf(subs={z: x})  # e(x)
        e_1_x = e_1_z.evalf(subs={z: x})  # e(1/x)

        # note that c_x and d_x are switched.
        a_x = polyval_dict(a, x)  # a(x)
        b_x = polyval_dict(b, x)  # b(x)
        c_x = (e_x - e_1_x) * sqrt(alpha) / mpc(2j)  # c(x)
        d_x = (e_x + e_1_x) * sqrt(alpha) / mpf(2)  # d(x)

        F.append(a_x * Id + b_x * I * X + c_x * I * Y + d_x * I * Z)

    return F


def step4(f, n):
    """
    Computes the FFT of f and returns a dict, c, of matrices.
    :param f: (list of 2 x 2 Sympy Matrices) the values of F(x), computed over a set of inputs in step3()
    :param n: (int) max. degree of a(z) or b(z) (computed in step_1())
    :return: c (dict of 2 x 2 mpmath matrices): the coefficients of F(z)
    """
    # Take the FFT of the f-values to find the matrices C.
    c_list = fft(f, dps=mp.dps)  # the C matrices, the Laurent coeffs of F(z)

    # Invert c_list: fft returns the matrix C_j as element c_list[-j]. We don't know why.
    c_list = [c_list[-i] for i in range(len(c_list))]
    c_list = [(matrix(c_list[i].evalf())) for i in range(len(c_list))]  # convert to mpmath matrices

    # Convert c_list to a dict with appropriate indexing
    d = len(f)
    c = defaultdict(lambda: zeros(2))  # stores the entries of c_list as a dict

    for i in range(-n, 0):  # {-n, -n + 1, ..., -1}
        c[2 * i] = c_list[i + d] / d  # set the coeffs of the negative-degree terms
    for i in range(n + 1):  # {0, 1, ..., n}
        c[2 * i] = c_list[i] / d  # set the coeffs of the positive-degree terms

    return c


def step5(n, c_matrices, return_matrices=False):
    """
    Computes E(t)s from C(t)s by Qs and Ps
    :param n: the same n output by step 1
    :param c_matrices: a dictionary of the 2n+1 coefficient matrices from the FFT of step 4. dict keys are -2n,-2n+2,...,2n
    :return: a dictionary of projector matrices (each entry is a tuple of P and Q, except E0) or a dictionary of angles (default)
    """

    big_es = {}  # key=superscript of each primitive matrix, currently represented as (P, Q) tuple. Except E_0 = C_0
    all_cs = {2 * n: c_matrices}
    for m in range(2 * n, 0, -1):  # m = 2n, 2n-1, ... 2, 1
        # (part i)
        c_m = all_cs[m][m]
        q_m = c_m.H * c_m

        spec_norm = sqrt(max(chop(eig(q_m.H * q_m)[0])))  # default tolerance of mpmath.chop
        q_m = q_m / spec_norm  # normalizing
        p_m = eye(2) - q_m

        # E = tP + t^(-1)Q , but really only P is needed
        # the matrices E are just a tuple of P and Q, except Esub0 is a single matrix
        big_es[m] = (p_m, q_m)

        # (part ii)
        all_cs[m - 1] = {}  # adding nested dictionary to dictionary
        for k in range(-m + 1, m, 2):
            all_cs[m - 1][k] = matrix(all_cs[m][k - 1]) * p_m + matrix(
                    all_cs[m][k + 1]) * q_m  # eq (18)
            if m == 1 and k == 0:  # add this last one ( C_0^(0) )
                big_es[0] = all_cs[0][0]

    if return_matrices:
        return big_es
    else:
        angle_decomposition = {}
        for E in big_es.keys():
            if E != 0:
                # finds the primitive matrix E(m),
                # takes P from the tuple, takes the upper right element,
                # then finds the corresponding angle
                # this works because eq(10) becomes Pj = ([0.5, 0.5exp(i*phi_j)], [0.5exp(-i*phi_j), 0.5]) Eq(10)
                angle_decomposition[E] = polar(big_es[E][0][0, 1])[1]
                if angle_decomposition[E] < 0:
                    angle_decomposition[E] = angle_decomposition[E] + 2 * pi
            else:
                angle_decomposition[0] = 2 * polar(big_es[0][0, 0])[
                    1]  # the polar angle of the first element of C_0
        return angle_decomposition



import argparse
import json
import logging
import os
import sys
from itertools import combinations
from math import ceil
# https://stackoverflow.com/a/25823885
from timeit import default_timer as timer
from typing import Iterable

from mpmath import mp, ln, nstr, mpf, arange, e, ldexp
from numpy import logspace, linspace

from qspd import jacobi_anger, step1, step2, step3, step4, step5
from qspd.json_encoding import DataEncoder
from qspd.tests import estimate_error


def collect_data(taus: Iterable = range(1, 6), epsilons=None, root_partitioner=None, big_ns=None):
    if epsilons is None:
        epsilons = [1e-4]
    if big_ns is None:
        big_ns = [None]
    if root_partitioner is None:
        def root_partitioner(big_r):
            num_real_roots = len(big_r)
            half_num_real_roots, remainder = divmod(num_real_roots, 2)
            combos = list(combinations(range(num_real_roots), half_num_real_roots))
            if remainder != 0:
                combos.extend(combinations(range(num_real_roots), half_num_real_roots + 1))
            for combo in combos:
                big_r_1_indices = set(combo)
                big_r_1 = [big_r[i] for i in combo]
                big_r_2 = [big_r[i] for i in range(num_real_roots) if i not in big_r_1_indices]
                yield big_r_1, big_r_2
    print("[")
    first = True
    for tau in taus:
        for epsilon in epsilons:
            for forced_big_n in big_ns:
                for record in create_records(tau, epsilon, root_partitioner, forced_big_n):
                    if not first:
                        print(",")
                    json.dump(record, sys.stdout, cls=DataEncoder)
                    sys.stdout.flush()
                    first = False
    print("\n]")


def create_records(tau, epsilon, root_partitioner, forced_n):
    logger = logging.getLogger("datacollection")
    logger.info(f"epsilon={nstr(epsilon)} tau={nstr(tau)} N={nstr(forced_n)}")
    early_times = {}

    start = timer()
    big_a, big_b = jacobi_anger(tau, epsilon, forced_n)
    big_n = max(big_a.degree, big_b.degree) if forced_n is None else forced_n
    early_times["step0"] = timer() - start

    mp.dps = int(ceil((big_n / 3) * ln(big_n / epsilon)))

    start = timer()
    a, b, n = step1(big_a, big_b, epsilon, forced_n)
    early_times["step1"] = timer() - start

    start = timer()
    p, roots = step2(a, b, n)
    early_times["step2"] = timer() - start

    big_c = [r for r in roots if r.imag > 0 and abs(r) < 1]
    big_r = [r.real for r in roots if r.imag == 0 and abs(r) < 1]
    assert 4 * len(big_c) + 2 * len(big_r) == len(roots)

    for partition in root_partitioner(big_r):
        big_r_1, big_r_2 = partition
        times = dict(early_times)

        start = timer()
        f = step3(a, b, n, p, roots, big_r_1=big_r_1, big_r_2=big_r_2)
        times["step3"] = timer() - start

        start = timer()
        c_mat_dict = step4(f, n)
        times["step4"] = timer() - start

        start = timer()
        angles = step5(n, c_mat_dict)

        times["step5"] = timer() - start

        error = estimate_error(big_a, big_b, angles)

        times["total"] = sum(time for step, time in times.items() if step != "step0")

        record = {
            "tau": tau,
            "epsilon": epsilon,
            "bigA": big_a,
            "bigB": big_b,
            "bigN": big_n,
            "dps": mp.dps,
            "a": a,
            "b": b,
            "n": n,
            "roots": roots,
            "bigR1": big_r_1,
            "bigR2": big_r_2,
            "bigF": f,
            "bigCs": c_mat_dict,
            "angles": angles,
            "error": error,
            "times": times
        }
        yield record


def main():
    parser = argparse.ArgumentParser(description="Collect data for different values of tau")
    parser.add_argument("--filename", type=str, default=None, help="The name of the file to save, or leave blank to print to stdout")
    parser.add_argument("--tau", type=float, nargs="+", help="The value for tau or the range (default=linear) of tau expressed by --tau min max numpoints (np.linspace)")
    parser.add_argument("--log-tau", action="store_true", help="If set, tau arguments are interpreted as np.logspace arguments instead of np.linspace, --tau must have 3 args")
    parser.add_argument("--epsilon", type=float, nargs="+", help="The value(s) of epsilon, controlling precision, or a logarithmic (base 10) range of epsilon expressed by --epsilon minexp maxexp numpoints")
    parser.add_argument("--n", type=str, nargs="+", help="The value of N=n or the linear range of n expressed by --n min max numpoints (np.linspace), uses minumum decomp if not set, uses minimum of biggest decomp for all iterations if set to auto")
    args = parser.parse_args()
    filename = args.filename
    old_out = sys.stdout
    if filename is not None:
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        sys.stdout = open(filename, mode="w")
    logger = logging.getLogger("datacollection")
    logger.addHandler(logging.StreamHandler())
    logger.setLevel(logging.DEBUG)

    def root_partitioner(big_r):
        num_real_roots = len(big_r)
        half_num_real_roots, remainder = divmod(num_real_roots, 2)
        combos = list(combinations(range(num_real_roots), half_num_real_roots))
        num_combos = len(combos)
        assert num_combos > 0
        if num_combos <= 1:
            combo = combos[0]
        else:
            combo = max(combos, key=lambda c: sum(big_r[i] for i in c))
            if num_combos > 2:
                logger.warning(f"Lots of combos: {combos}")
        big_r_1_indices = set(combo)
        big_r_1 = [big_r[i] for i in combo]
        big_r_2 = [big_r[i] for i in range(num_real_roots) if i not in big_r_1_indices]
        yield big_r_1, big_r_2

    tau = args.tau
    if tau is None:
        taus = [10]
    elif len(tau) == 1:
        taus = tau
    elif len(tau) == 3:
        if args.log_tau:
            taus = logspace(tau[0], tau[1], int(tau[2]))
        else:
            taus = linspace(tau[0], tau[1], int(tau[2]))
    else:
        raise AssertionError("--tau should have 1 (ex: 80), or 3 (ex: 10 20 5) (np.linspace) arguments")
    min_str = nstr(min(taus))
    max_str = nstr(max(taus))
    if args.log_tau and len(taus) == 3:
        min_str = "10^" + min_str
        max_str = "10^" + max_str

    epsilon = args.epsilon
    if epsilon is None:
        epsilons = [0.0001]
    elif len(epsilon) == 1:
        epsilons = epsilon
    elif len(epsilon) == 3:
        epsilons = logspace(epsilon[0], epsilon[1], epsilon[2])
    else:
        raise AssertionError("--epsilon should have 1 (ex: .000001), or 3 (np.logspace, ex: -4 -9 4) arguments")

    n = args.n
    if n is None:
        ns = n
    elif n == ["auto"]:
        ns = [int(ceil(ldexp(e, -1) * max(taus) - ln(min(epsilons))))]
    elif len(n) == 1:
        ns = [int(n[0])]
    elif len(n) == 3:
        ns = [int(round(i)) for i in linspace(*list(map(int, n)))]
    else:
        raise AssertionError("--n should have 1 (ex: 145), or 3 (ex: 30, 36, 1) (np.linspace) integer arguments")

    if n is None:
        logger.info(f"Running {len(taus)} iteration(s) from tau={min_str} to tau={max_str}, {len(epsilons)} iteration(s) from epsilon={min(epsilons)} to epsilon={max(epsilons)}, and N=n is not forced larger than needed ")
    else:
        logger.info(f"Running {len(taus)} iteration(s) from tau={min_str} to tau={max_str}, {len(epsilons)} iteration(s) from epsilon={min(epsilons)} to epsilon={max(epsilons)}, and {len(ns)} iteration(s) from N=n={min(ns)} to N=n={max(ns)} ")
    collect_data(taus=taus, epsilons=epsilons, root_partitioner=root_partitioner, big_ns=ns)

    if filename is not None:
        sys.stdout.close()
        sys.stdout = old_out


if __name__ == '__main__':
    main()



import argparse
import json
import logging
from collections import defaultdict

import numpy as np
from matplotlib import pyplot as plt
from mpmath import log, nint

from qspd.json_encoding import decode_data


def plot(data, light=False, image_file=None, image_size=(640, 480), line_colors=None, best_fit_start=None):
    logger = logging.getLogger(__name__)
    logger.addHandler(logging.StreamHandler())
    logger.setLevel(logging.DEBUG)

    if not light:
        plt.style.use("dark_background")
    brightness = 1 if light else 0
    alpha = 1 if image_file is None else 0
    face_color = (brightness, brightness, brightness, alpha)
    width, height = image_size
    fig = plt.figure(facecolor=face_color, figsize=(width / 100, height / 100))
    ax = fig.add_subplot(111)
    ax.set_facecolor((0, 0, 0, 0))
    plt.grid()

    time_dict = defaultdict(dict)
    for record in data:
        time_dict[record["error"]][record["tau"]] = record["times"]["total"]

    for color_index, (error, tau_time) in enumerate(reversed(list(time_dict.items()))):
        taus, times = zip(*tau_time.items())
        color = line_colors[color_index % len(line_colors)] if line_colors is not None else None
        exponent = int(nint(log(error, b=10)))
        plt.scatter(taus, times, label=f"error $= 10^{{{exponent}}}$", color=color)
        if best_fit_start is not None:
            tau_indices = [i for i, tau in enumerate(taus) if taus[i] >= best_fit_start]
            large_taus = np.array([float(taus[i]) for i in tau_indices])
            large_times = np.array([times[i] for i in tau_indices])
            log_taus = np.log10(large_taus)
            log_times = np.log10(np.array(large_times))
            fit = np.polyfit(log_taus, log_times, 1)
            poly = np.poly1d(fit)
            plt.plot(large_taus, 10 ** poly(log_taus), color=color, linestyle=":")
    plt.legend()
    plt.ylabel("running time (s)")
    plt.xlabel(r"$\tau$")
    plt.xscale("log")
    plt.yscale("log")
    if image_file is None:
        plt.show()
    else:
        plt.savefig(image_file, transparent=True)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
            description="Plot runtime as a function of tau")
    parser.add_argument("files", metavar="file", type=str, nargs="+",
                        help="The JSON file(s) to get data from")
    parser.add_argument("--best-fit-start", type=float,
                        help="Adds a best-fit line to all points after (including) the given tau")
    parser.add_argument("--line-colors", type=str, nargs="+",
                        help="The colors of the plotted lines")
    parser.add_argument("--light", action="store_true", help="Add a light background")
    parser.add_argument("--image-file", type=str, help="The name of the image file to save")
    parser.add_argument("--image-size", type=int, nargs=2, default=[640, 480],
                        help="The width and height of the image")
    args = parser.parse_args()

    data = []
    for filename in args.files:
        with open(filename) as file:
            data.extend(decode_data(json.load(file)))
    image_size_list = args.image_size
    image_size = (image_size_list[0], image_size_list[1])

    plot(data, light=args.light, image_file=args.image_file, image_size=image_size,
         line_colors=args.line_colors, best_fit_start=args.best_fit_start)



from enum import Enum
from fractions import Fraction
from numbers import Rational

from mpmath import mpf, mpc, fadd, fsub, fmul, fdiv, fraction, fsum


class ImaginaryRational:
    """
    A rational number multiplied by i. Squaring an ImaginaryRational will yield a negative Fraction.
    """

    def __new__(cls, numerator=0, denominator=None):
        self = super(ImaginaryRational, cls).__new__(cls)
        self._value = Fraction(numerator, denominator)
        return self

    @property
    def value(self):
        return self._value

    @property
    def numerator(self):
        return self._value.numerator

    @property
    def denominator(self):
        return self._value.denominator

    # repr(self)
    def __repr__(self):
        return f"ImaginaryRational({self.numerator}, {self.denominator})"

    # str(self)
    def __str__(self):
        return f"{self.numerator}i/{self.denominator}"

    # self + other
    def __add__(self, other):
        if other == 0:
            return self
        if isinstance(other, ImaginaryRational):
            return ImaginaryRational(self.value + other.value)
        elif isinstance(other, (mpf, mpc)):
            return fadd(self.mpc, other)
        if isinstance(other, Rational) and self.value == 0:
            return other
        return NotImplemented

    # other + self
    def __radd__(self, other):
        return self + other

    # self - other
    def __sub__(self, other):
        if other == 0:
            return self
        if isinstance(other, ImaginaryRational):
            return ImaginaryRational(self.value - other.value)
        elif isinstance(other, (mpf, mpc)):
            return fsub(self.mpc, other)
        if isinstance(other, Rational) and self.value == 0:
            return other
        return NotImplemented

    # other - self
    def __rsub__(self, other):
        if other == 0:
            return ImaginaryRational(-self.value)
        if isinstance(other, ImaginaryRational):
            return ImaginaryRational(other.value - self.value)
        elif isinstance(other, (mpf, mpc)):
            return fsub(other, self.mpc)
        if isinstance(other, Rational) and self.value == 0:
            return other
        return NotImplemented

    # self * other
    def __mul__(self, other):
        if other == 0:
            return 0
        if isinstance(other, Rational):
            return ImaginaryRational(self.value * other)
        elif isinstance(other, ImaginaryRational):
            return Fraction(-self.value * other.value)
        elif isinstance(other, (mpf, mpc)):
            return fmul(other, self.mpc)
        return NotImplemented

    # other * self
    def __rmul__(self, other):
        return self * other

    # self / other
    def __truediv__(self, other):
        if isinstance(other, Rational):
            return ImaginaryRational(self.value / other)
        if isinstance(other, ImaginaryRational):
            return Fraction(self.value / other.value)
        elif isinstance(other, (mpf, mpc)):
            return fdiv(self.mpc, other)
        return NotImplemented

    # other / self
    def __rtruediv__(self, other):
        if isinstance(other, Rational):
            return ImaginaryRational(-other / self.value)
        elif isinstance(other, ImaginaryRational):
            return Fraction(other.value / self.value)
        elif isinstance(other, (mpf, mpc)):
            return fdiv(other, self.mpc)
        return NotImplemented

    # self ** power
    def __pow__(self, power):
        if not isinstance(power, int):
            return NotImplemented
        result = self.value ** power
        power_mod_4 = power % 4
        if power_mod_4 == 0:
            return Fraction(result)
        elif power_mod_4 == 1:
            return ImaginaryRational(result)
        elif power_mod_4 == 2:
            return Fraction(-result)
        else:
            assert power_mod_4 == 3
            return ImaginaryRational(-result)

    # +self
    def __pos__(self):
        return self

    # -self
    def __neg__(self):
        return ImaginaryRational(-self.value)

    # abs(self)
    def __abs__(self):
        return abs(self.value)

    # hash(self)
    def __hash__(self):
        value = self.value
        if value == 0:
            return hash(0)
        return 13 + hash(self.value)

    # self == other
    def __eq__(self, other):
        value = self.value
        if value == 0 and other == 0:
            return True
        if not isinstance(other, ImaginaryRational):
            return False
        return value == other.value

    def __bool__(self):
        return self.value != 0

    @property
    def mpc(self):
        return mpc(0, fraction(self.numerator, self.denominator))


class Parity(Enum):
    """
    The parity of a function, either even or odd.
    """
    ODD = 1
    EVEN = 2


# TODO refactor InputFunction
# a function, with definite parity and coefficients accessible via indexing
class InputFunction:
    """
    A function that can be input to Haah's algorithm. This function can be thought of as a
    2pi-periodic function applied to real numbers, represented in terms of its Fourier coefficients,
    or it can be thought of as a Laurent polynomial
    """
    def __init__(self, coefficients, parity, degree=None):
        self._coefficients = coefficients
        self.parity = parity
        if degree is None:
            powers = (abs(power) for power, coefficient in coefficients.items() if coefficient != 0)
            self.degree = max(powers, default=0)
        else:
            self.degree = degree

    def __getitem__(self, item):
        return self._coefficients.get(item, mpf(0))

    def __iter__(self):
        return iter(range(-self.degree, self.degree + 1))

    def __str__(self):
        return str(dict(self._coefficients))

    def __repr__(self):
        return repr(dict(self._coefficients))


def rational_to_mpc(x):
    """
    Produces an mpmath floating-point number with a zero real part and an imaginary part
    approximately equal to the value of the parameter

    :param x: a number, either Rational or ImaginaryRational
    :return: an mpf or mpc approximation to the parameter
    """
    if isinstance(x, Rational):
        return fraction(x.numerator, x.denominator)
    elif isinstance(x, ImaginaryRational):
        return x.mpc
    raise TypeError


def polyval_dict(coeffs, x):
    """
    Evaluates a Laurent polynomial at a particular value. This function can be represented as a
    dictionary with rational or floating-point coefficients, or as a FloatFunction.

    :param coeffs: the coefficients, indexed by power
    :param x: an mpmath.mpc complex value at which to evaluate the function
    :return:
    """
    def eval_term(power):
        coeff = coeffs[power]
        if isinstance(coeff, (Rational, ImaginaryRational)):
            return rational_to_mpc(coeff) * x ** power
        else:
            return coeff * x ** power

    return fsum(eval_term(power) for power in coeffs)



import argparse
import json
import logging
import math
from collections import defaultdict

import matplotlib.pyplot as plt
from mpmath import pi, ceil, floor
from sympy import fft

from qspd.json_encoding import decode_data


def visualize(data, angle_indices=None, fourier=False, center_pi=False, epsilon=False,
        light=False, transparent=False, image_file=None, image_size=(640, 480), marker=False, line_colors=None,
        pi_x=False, derivative=False):
    logger = logging.getLogger(__name__)
    logger.addHandler(logging.StreamHandler())
    logger.setLevel(logging.DEBUG)

    if not light:
        plt.style.use("dark_background")
    brightness = 1 if light else 0
    alpha = 1 if image_file is None else 0
    face_color = (brightness, brightness, brightness, alpha)
    width, height = image_size
    fig = plt.figure(facecolor=face_color, figsize=(width / 100, height / 100))
    ax = fig.add_subplot(111)
    ax.set_facecolor((0, 0, 0, 0))
    plt.grid()

    angle_dict = defaultdict(dict)
    x_handle = "epsilon" if epsilon else "tau"
    for record in data:
        angles = record["angles"]
        for angle_index, angle in enumerate(angles):
            x_val = record[x_handle]
            angle_dict[angle_index][x_val] = float(angle)
    if angle_indices is None or len(angle_indices) == 0:
        angle_indices = angle_dict.keys()
    for color_index, angle_index in enumerate(angle_indices):
        x_vals, y_vals = zip(*sorted(angle_dict[angle_index].items()))
        color = line_colors[color_index % len(line_colors)] if line_colors is not None else None
        if fourier:
            if center_pi:
                y_vals = [y_val - pi for y_val in y_vals]
            y_vals = [float(abs(coeff)) for coeff in fft(y_vals, dps=15)]
            x_vals = range(len(y_vals))
        elif derivative:
            y_vals = [(y_vals[i + 1] - y_vals[i]) / (x_vals[i + 1] - x_vals[i]) for i in
                      range(len(y_vals) - 1)]
            x_vals = [(x_vals[i] + x_vals[i + 1]) / 2 for i in range(len(x_vals) - 1)]
        marker_pattern = "." if marker else None
        plt.plot(x_vals, y_vals, marker=marker_pattern, color=color, label=angle_index)

    if fourier:
        plt.xticks([])
        plt.yticks([])
    else:
        if epsilon:
            plt.xlabel(r"$\epsilon$")
            plt.xscale("log")
        else:
            plt.xlabel(r"$\tau$")

            if pi_x:
                min_tau = min(record["tau"] for record in data)
                min_tick_over_2pi = int(floor(min_tau / 2 * pi))

                max_tau = max(record["tau"] for record in data)
                ticks_over_2pi = list(range(min_tick_over_2pi, int(ceil(max_tau / (2 * pi)))))
                ticks = [tick_over_2pi * 2 * math.pi for tick_over_2pi in ticks_over_2pi]

                def label_tick(tick_over_2pi):
                    if tick_over_2pi == 0:
                        return "0"
                    else:
                        return f"${tick_over_2pi * 2}\\pi$"

                labels = [label_tick(tick_over_2pi) for tick_over_2pi in ticks_over_2pi]
                plt.xticks(ticks, labels)
        if derivative:
            y_label = "derivative of angle value"
        else:
            y_label = "angle value"
            plt.ylim(0, 2 * math.pi)
            y_labels = ["0", r"$\frac{\pi}{2}$", r"$\pi$", r"$\frac{3\pi}{2}$", r"$2\pi$"]
            plt.yticks([0, math.pi / 2, math.pi, 3 * math.pi / 2, 2 * math.pi], y_labels)
        plt.ylabel(y_label)

    plt.legend()
    if image_file is not None:
        plt.savefig(image_file, transparent=transparent)
    else:
        plt.show()


def main():
    parser = argparse.ArgumentParser(
            description="Visualize line plots of angles for different values of tau")
    parser.add_argument("files", metavar="file", type=str, nargs="+",
                        help="The JSON file(s) to get data from")
    parser.add_argument("--angle-indices", metavar="angle-index", type=int, nargs="+",
                        help="Specific angle indices to plot")
    parser.add_argument("--epsilon", action="store_true",
                        help="Plot angles changing with respect to epsilon, rather than tau")
    parser.add_argument("--derivative", action="store_true",
                        help="take the derivative of the angle value")
    parser.add_argument("--fourier", action="store_true",
                        help="Take the Fourier transform of the desired angle indices")
    parser.add_argument("--center-pi", action="store_true",
                        help="Subtract pi from all angles before taking the Fourier transform")
    parser.add_argument("--pi-x", action="store_true", help="x axis in terms of pi")
    parser.add_argument("--marker", action="store_true",
                        help="Display circles at each point along the plot")
    parser.add_argument("--line-colors", type=str, nargs="+",
                        help="The colors of the plotted lines")
    parser.add_argument("--light", action="store_true", help="Add a light background")
    parser.add_argument("--transparent", action="store_true", help="Make background transparent")
    parser.add_argument("--image-file", type=str, help="The name of the image file to save")
    parser.add_argument("--image-size", type=int, nargs=2, default=[640, 480],
                        help="The width and height of the image")
    args = parser.parse_args()

    data = []
    for filename in args.files:
        with open(filename) as file:
            data.extend(decode_data(json.load(file)))
    image_size_list = args.image_size
    image_size = (image_size_list[0], image_size_list[1])

    visualize(data, angle_indices=args.angle_indices, fourier=args.fourier,
              center_pi=args.center_pi, epsilon=args.epsilon, light=args.light, transparent=args.transparent,
              image_file=args.image_file, image_size=image_size, marker=args.marker,
              line_colors=args.line_colors, pi_x=args.pi_x, derivative=args.derivative)


if __name__ == '__main__':
    main()



import argparse
import json
import logging

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import Normalize, ListedColormap
from mpmath import nstr

from qspd.json_encoding import decode_data


def visualize(data, epsilon_scale=False, left_aligned=False, auto_scale=False, n_scale=False,
        evens_only=False, odds_only=False, light=False, image_file=None, image_size=(640, 480),
        no_axes=False):
    logger = logging.getLogger("colorviz")
    logger.addHandler(logging.StreamHandler())
    logger.setLevel(logging.DEBUG)

    if not light:
        plt.style.use("dark_background")
    brightness = 1 if light else 0
    alpha = 1 if image_file is None else 0
    face_color = (brightness, brightness, brightness, alpha)
    width, height = image_size
    fig = plt.figure(facecolor=face_color, figsize=(width / 100, height / 100))
    ax = fig.add_subplot(111)
    if no_axes:
        fig.frameon = False
        ax.set_axis_off()
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)

    if evens_only:
        data = [record for record in data if record["bigN"] % 2 == 0]
    elif odds_only:
        data = [record for record in data if record["bigN"] % 2 != 1]

    colormap = plt.get_cmap("hsv")
    ns = [record["n"] for record in data]
    if ns[1:] != ns[:-1]:  # https://stackoverflow.com/a/3844832
        # different values of n
        colormap_data = plt.get_cmap("hsv")(np.arange(256))
        colormap_data[0, :] = np.array([0, 0, 0, 0])  # zero is black
        colormap = ListedColormap(colormap_data)

    def add_padding(angles):
        padding = (max_num_angles - len(angles)) // 2
        if left_aligned:
            return np.pad(angles, (0, padding * 2), "constant")
        else:
            center = np.pad(angles[1:], (padding, padding), "constant")
            return np.insert(center, 0, np.array(angles)[0:1])

    angle_lists = [[float(angle) for angle in record["angles"]] for record in data]
    taus = [record["tau"] for record in data]
    epsilons = [record["epsilon"] for record in data]
    angle_lists.reverse()
    taus.reverse()
    epsilons.reverse()
    ns.reverse()
    max_num_angles = max(len(angles) for angles in angle_lists)
    padded = [add_padding(angles) for angles in angle_lists]
    all_angles = np.stack(padded)
    all_angles[(0 < all_angles) & (all_angles < 1 / 40)] = 2 * np.pi

    tau_increment = data[1]["tau"] - data[0]["tau"] if len(data) > 1 else 1
    if auto_scale and not epsilon_scale:
        min_tau = float(min(taus)) - tau_increment / 2
        max_tau = float(max(taus)) + tau_increment / 2
        extent = [-0.5, max_num_angles - 0.5, min_tau, max_tau]
    else:
        extent = None
    norm = Normalize(0, 2 * np.pi)
    plt.imshow(all_angles, norm=norm, aspect="auto", cmap=colormap, extent=extent)

    if no_axes:
        fig.subplots_adjust(bottom=0)
        fig.subplots_adjust(top=1)
        fig.subplots_adjust(right=1)
        fig.subplots_adjust(left=0)
    else:
        plt.xlabel("angle index")
        if n_scale:
            plt.ylabel("N").set_rotation(0)
            super_title = "Changing N, constant Tau = " + str(data[0]["tau"]) + " epsilon = " + str(
                    data[0]["epsilon"])
            plt.suptitle(super_title, fontsize=14, fontweight='bold')

        elif epsilon_scale:
            plt.ylabel(r"$\epsilon$").set_rotation(0)
        else:
            plt.ylabel(r"$\tau$").set_rotation(0)

        if not auto_scale:
            if n_scale:
                plt.yticks(range(len(ns)), [int(n) for n in ns])
            elif epsilon_scale:
                # TODO config scale ticks
                ticks = [tick for tick in range(len(epsilons)) if tick % 20 == 0]
                labels = [epsilons[tick] for tick in ticks]
                plt.yticks(ticks, labels)
            else:
                plt.yticks(range(len(taus)), [nstr(tau) for tau in taus])

    if image_file is None:
        plt.show()
    else:
        pad_inches = 0 if no_axes else None
        plt.savefig(image_file, transparent=True, pad_inches=pad_inches)


def main():
    parser = argparse.ArgumentParser(
            description="Visualize angles as colors for different values of tau")
    parser.add_argument("files", metavar="file", type=str, nargs="+",
                        help="The JSON file(s) to get data from")
    parser.add_argument("--epsilon-scale", action="store_true",
                        help="Add if epsilon is changing rather than tau")
    parser.add_argument("--auto-scale", action="store_true",
                        help="Add if there are no duplicates and there shouldn't be tick marks for all values of tau")
    parser.add_argument("--n-scale", action="store_true", help="Add if y axis should be N sweep")
    parser.add_argument("--left-aligned", action="store_true",
                        help="Add if colors should be left-aligned (default is centered)")
    parser.add_argument("--evens-only", action="store_true",
                        help="Add if you only want to plot even ns")
    parser.add_argument("--odds-only", action="store_true",
                        help="Add if you only want to plot odd ns")
    parser.add_argument("--no-axes", action="store_true",
                        help="Add if you don't want any axes, meaning that the image will be only the colors")
    parser.add_argument("--light", action="store_true", help="Add a light background")
    parser.add_argument("--image-file", type=str, help="The name of the image file to save")
    parser.add_argument("--image-size", type=int, nargs=2, default=[640, 480],
                        help="The width and height of the image")

    args = parser.parse_args()

    data = []
    for filename in args.files:
        with open(filename) as file:
            data.extend(decode_data(json.load(file)))
    image_size_list = args.image_size
    image_size = (image_size_list[0], image_size_list[1])

    visualize(data, epsilon_scale=args.epsilon_scale, left_aligned=args.left_aligned,
              auto_scale=args.auto_scale, n_scale=args.n_scale, evens_only=args.evens_only,
              odds_only=args.odds_only, light=args.light, image_file=args.image_file,
              image_size=image_size, no_axes=args.no_axes)


if __name__ == '__main__':
    main()



import unittest
from collections import defaultdict
from fractions import Fraction
from random import randint

from mpmath import fabs, ldexp, matrix, norm, ln, ceil
from mpmath import mp, rand, erfinv, fdiv, fmul, fsub, almosteq, nstr
from mpmath import mpc, expj, mpf, pi
from sympy import Matrix, zeros

from qspd import step1, InputFunction, Parity, step2, step4, step3
from qspd.tests import recompose
from qspd.util import rational_to_mpc, polyval_dict, ImaginaryRational


# Tests steps 1 - 5 individually or in small groups
# Run this script to perform unit testing.
class UnitTesting(unittest.TestCase):
    def test_step1(self):
        for j in range(50):
            mp.dps = 150
            big_n = randint(1, 20)
            big_a, big_b = defaultdict(lambda: mpf(0)), defaultdict(lambda: mpc(0))
            for i in range(big_n + 1):
                # citation:
                # https://docs.sympy.org/0.7.1/modules/mpmath/functions/expintegrals.html#erfinv
                a_i, b_i = erfinv(2 * rand() - 1), mpc(0, erfinv(2 * rand() - 1))
                big_a[i], big_a[-i] = a_i, a_i
                if i != 0:
                    big_b[i], big_b[-i] = b_i, -b_i
            epsilon = mpf(10) ** randint(-10, -3)
            epsilon_over_n = fdiv(epsilon, big_n, rounding='u')
            one_minus_five_epsilon = fsub(1, fmul(5, epsilon, exact=True), exact=True)
            big_a = InputFunction(big_a, Parity.EVEN, big_n)
            big_b = InputFunction(big_b, Parity.ODD, big_n)
            a, b, n = step1(big_a, big_b, epsilon)

            mp.dps = 150

            def test_coefficient(big_p, p, i):
                expected = fmul(one_minus_five_epsilon, big_p[i], exact=True)
                actual = rational_to_mpc(p[i])
                are_equal = almosteq(expected, actual, abs_eps=epsilon_over_n)
                if not are_equal:
                    print(f"expected: {nstr(expected, n=10)} but was: {nstr(actual, n=10)}")
                self.assertTrue(are_equal)

            for i in range(n):
                test_coefficient(big_a, a, i)
                test_coefficient(big_b, b, i)
                self.assertEqual(a[i], a[-i])
                self.assertEqual(b[i], -b[-i])

    def test_step2(self):
        # Step 1: Construct the inputs to step2()
        n = 5
        mp.dps = 150

        def im(x):
            return ImaginaryRational(x)

        a_coeffs = [(-5, 1), (-4, 1), (-3, 1), (-2, 4), (-1, 5), (0, Fraction(6, 5)), (1, 1),
                    (2, 4), (3, -4),
                    (4, -1), (5, 0)]  # a has real coeffs
        b_coeffs = [(-5, im(0)), (-4, im(1)), (-3, im(2)), (-2, im(3)), (-1, im(4)), (0, im(5)),
                    (1, im(-4)),
                    (2, im(-3)), (3, im(-3)), (4, im(-1)), (5, im(0))]  # b has imaginary coeffs

        # construct a dict from the coeffs in a_list and b_list
        a = defaultdict(lambda: 0)
        b = defaultdict(lambda: 0)
        for key, value in a_coeffs:
            a[key] = value
        for key, value in b_coeffs:
            b[key] = value

        # Outputs of step2()
        p, roots = step2(a, b, n)

        # Step 3: Check that p is the correct polynomial
        radius = 3  # inputs range from -radius to radius
        step_size = mpf(1) / 5  # number of points along one coordinate from 0 to radius

        num = int(radius / step_size)
        for alpha in range(-num, num + 1):
            for beta in range(-num, num + 1):
                z = mpc(alpha, beta) * step_size  # the complex value input into a, b, and p
                if z == mpc(0, 0):
                    continue

                a_z = polyval_dict(a, z)  # the value of a evaluated at z
                b_z = polyval_dict(b, z)
                p_z = polyval_dict(p, z)

                err = fabs(p_z - (1 - a_z ** 2 - b_z ** 2))  # absolute error
                self.assertLessEqual(err, ldexp(1, -15))  # err <= 2^-15

        # Step 4: Check that the roots are correct
        for r in roots:
            err = fabs(polyval_dict(p, r))  # absolute error
            self.assertLessEqual(err, ldexp(1, -15))  # err <= 2^-15
        self.assertTrue(True)  # does nothing

    def test_step4(self):
        # Step 1: Define a Laurent polynomial whose coeffs are 2x2 matrices
        coeffs = defaultdict(lambda: zeros(2))  # the coeffs of a Laurent polynomial
        coeffs[-1] = Matrix([[mpc(0, 1), 0], [mpc(2, .3), 1]])
        coeffs[0] = Matrix([[1, 0], [0, 1]])
        coeffs[1] = Matrix([[1, 0], [0, 2]])
        coeffs[2] = Matrix([[mpc(1, 2), 1], [mpc(4, 1.7), 0]])

        # Step 2: Evaluate the polynomial on at least 2n+1 inputs
        n = max(abs(j) for j in coeffs.keys())  # the degree of the Laurent polynomial
        d = 2 ** ((2 * n + 1).bit_length())  # number of inputs
        F = []  # the value of the polynomial at each input

        for k in range(d):
            z = expj(2 * pi * mpf(k) / d)  # the input value
            F_z = zeros(2)  # the function's output
            for power in coeffs:
                F_z += coeffs[power] * (z ** power)
            F.append(F_z)

        # Step 3: Recover the coeffs by taking the FFT of vals
        fft_coeffs = step4(F, n)

        # Check that fft_coeffs are close to coeffs
        for i in fft_coeffs.keys():
            err = norm(fft_coeffs[i] - matrix(coeffs[i / 2]))
            self.assertLessEqual(err, ldexp(1, -15))  # err <= 2^-15

    def test_step3_4(self):
        """
        Test step4() using the output of step3().
        Treat the coeffs output by step4() as a function. Compute the value of this function for many inputs.
        The function values should match the values output by step3().
        """
        # Step 1: Perform steps 1-4 on random input (code taken from main.py)
        mp.dps = 150
        epsilon = 0.0001  # precision control, epsilon should be between 0 and 1
        angles, big_a, big_b = _recompose_random(2)
        a, b, n = step1(big_a, big_b, epsilon)
        p, roots = step2(a, b, n)
        F = step3(a, b, n, p, roots)
        fft_coeffs = step4(F, n)

        # Step 2: Compute the values of function F using the coeffs given by step 4
        d = len(F)  # number of function values to compute
        F_est = []  # the estimated function values
        for k in range(d):
            z = expj(2 * pi * mpf(k) / mpf(d))  # the input value
            F_z = matrix(zeros(2))  # the function's output
            for power in fft_coeffs:
                F_z += fft_coeffs[power] * (z ** (power / 2))
            F_est.append(F_z)

        # Step 3: Compare the values in F and est_F
        for i in range(d):
            err = norm(F_est[i] - matrix(F[i].evalf()))
            self.assertLessEqual(err, ldexp(1, -15))  # err <= 2^-15

    def test_step1_2_3(self):
        """
        Check that steps 1-3 compute the correct function F.
        The input to step1() is a function, described by big_a and big_b: f(t) = big_a(t^2) + i * big_b(t^2).
        Compute the value of this function over many inputs.
        Then check that f(t^2) = <+|F(t)|+> for these inputs, where F(t) is the output of step3().
        """
        # Step 1: Generate an input to step1().
        epsilon = ldexp(1, -150)  # precision control, epsilon should be between 0 and 1
        angles, big_a, big_b = _recompose_random(3)
        big_n = max(big_a.degree, big_b.degree)
        mp.dps = int(ceil((big_n / 3) * ln(big_n / epsilon)))

        f = defaultdict()  # f(z) = big_a(z) + i * big_b(z)
        for i in range(-big_n, big_n + 1):
            f[i] = big_a[i] + mpc(0, 1) * big_b[i]

        # Step 2: Execute steps 1-3 to get function values of F(t)
        def sum_div_2(mat):
            """
            :param mat: a 2x2 matrix
            :return: the sum of the four elements of the matrix, divided by 2
            This is equivalent to: <+|mat|+>
            """
            return (mat[0, 0] + mat[0, 1] + mat[1, 0] + mat[1, 1]) / 2

        a, b, n = step1(big_a, big_b, epsilon)
        p, roots = step2(a, b, n)
        F = step3(a, b, n, p, roots)  # the coeffs of F(z)
        F1 = [sum_div_2(matrix(F[i])) for i in range(len(F))]  # F_1(z) = <+|F(z)|+>

        # Step 3: Check that <+|F(z)|+> = f(z)
        d = len(F1)  # number of inputs to test
        for k in range(d):
            z = expj(2 * pi * mpf(k) / d)  # the input value
            f_z = polyval_dict(f, z)  # f(z)

            err = fabs(F1[k] - f_z)  # /fabs(f_val)
            self.assertLessEqual(err, ldexp(1, -15))  # err <= 2^-15


def _recompose_random(n):
    """
    Performs reverse_decomp on a random set of 2n+1 angles
    :param n: the degree of the resulting function, A(z) + iB(z), assuming no terms cancel
    :return: the list of 2n+1 random angles, the functions A(z) and B(z), as FloatFunctions,
            and the degree, big_n, of the functions
    """
    angles = [rand() * 2 * pi for _ in range(2 * n + 1)]
    a, b = recompose(angles)
    return angles, a, b


if __name__ == '__main__':
    unittest.main()



import setuptools

long_description = """This package implements Jeongwan Haah's algorithm for quantum signal processing as described in this 
[paper](https://arxiv.org/abs/1806.10236). It includes a step by step 
algorithm that decomposes periodic functions (often from quantum signal processing) into a product of primitive matrices,
represented as a list of angles. The algorithmic complexity is O(N<sup>3</sup> polylog(N/𝜺)) where N is the degree of the
periodic function and 𝜺 is the precision parameter. The runtime bottleneck is a polynomial rootfinding in step 2. Haah 
concludes that the error is at most the error input (=15𝜺).
DISCLAIMER: This is in initial investigation stages, and the usage is subject to change. 
"""

setuptools.setup(
        name="qspd",
        version="1.0",
        author="Andrew Childs",
        author_email="amchilds@umd.edu",
        description="Quantum Signal Processing Decomposition",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://gitlab.umiacs.umd.edu/amchilds/qspd",
        packages=setuptools.find_packages(),
        install_requires=[
            "mpmath", "sympy", "numpy", "matplotlib"
        ],
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: OS Independent",
        ],
)



from collections import defaultdict

from mpmath import expj, matrix, eye, zeros, chop, mpc, pi, mpf, j, expjpi, workdps

from qspd import InputFunction, Parity
from qspd.util import polyval_dict


def estimate_error(big_a, big_b, angles, num_points=None):
    """
    Calculates the closeness between the input function A+Bi and the estimated version A'+B'i.
    We input values from the unit circle into both functions and compare the difference in outputs.
    The number of evenly spaced points is 5 times the polynomial degree

    :param big_a: Dictionary of coefficients of original A
    :param big_b: Dictionary of coefficients of original B
    :param angles: used to recompose the the A' + B' i
    :return: the maximum difference between the original A+iB and the resulting A'+B'i
    """
    num_points = int(5*(len(angles)-1)/2)
    max_error = 0
    output_big_a, output_big_b = recompose(angles)
    for k in range(num_points):
        z = expjpi(2*mpf(k)/mpf(num_points))
        expected = polyval_dict(big_a, z) + j*polyval_dict(big_b, z)
        calculated = polyval_dict(output_big_a, z) + j*polyval_dict(output_big_b, z)
        max_error = max(max_error, abs(expected-calculated))
    return max_error


def recompose(angles):
    """
    Reverses the matrix decomposition performed by steps 1-5:
    Given the angles output by step5(), this function computes A(z) and B(z), the inputs to step1().
    :param angles: (list of mpf's) the 2n+1 angles computed by step5().
                    Each angle[j] corresponds to a matrix E_j in equation 19.
                    len(angles) must be odd
    :return: The functions A(z) and B(z), as FloatFunctions,
            and the degree, big_n, of the functions
    """

    big_n, remainder = divmod(len(angles),
                              2)  # big_n is the degree of the polynomial we will compute
    if remainder == 0:  # there should be 2*big_n + 1 angles
        raise ValueError('angles must have an odd number of elements')

    # Step 1: Compute the matrix E0, corresponding to angle[0]
    e_0 = matrix([[expj(angles[0] / 2), 0],
                  [0, expj(-angles[0] / 2)]])  # the matrix E_0, defined in Haah section 3

    identity = eye(2)  # 2 x 2 identity matrix
    zero = zeros(2)  # 2 x 2 matrix of zeros

    prod = defaultdict(lambda: zero)  # the product E_0 * E_1 * E_2 * ... * E_{max_k}
    prod[0] = e_0

    # Step 2: Compute prod, the product of the E-matrices
    for k in range(1, len(angles)):  # omit angles[0]
        p_k = matrix([[1, expj(angles[k])],
                      [expj(-angles[k]), 1]]) / 2  # The projector matrix P corresponding to E_k
        q_k = identity - p_k  # The other projector Q corresponding to E_k

        # Each E_k_dagger = (P_k * t + Q_k * t^-1)^dagger is a binomial. Now expand the binomials into one polynomial.
        updated_prod = defaultdict(lambda: zero)
        for j, coeff in prod.items():
            updated_prod[j + 1] += coeff * q_k.H
            updated_prod[j - 1] += coeff * p_k.H
        prod = updated_prod  # the coefficients of prod are indexed by their t-degree, not z-degree

    # Step 3: Decompose each matrix in prod into real (A) and imaginary (B) matrices
    a = defaultdict(lambda: 0)  # the coeffs of the polynomial A(z)
    b = defaultdict(lambda: 0)  # the coeffs of the polynomial B(Z)

    def sum_div_2(mat):
        """
        :param mat: a 2x2 matrix
        :return: the sum of the four elements of the matrix, divided by 2
        This is equivalent to: <+|mat|+>
        """
        return (mat[0, 0] + mat[0, 1] + mat[1, 0] + mat[1, 1]) / 2

    for j in range(big_n + 1):
        # Note: We use prod[2 * j] to compute a[j], b[j] because a and b are indexed by z-degree,
        #       whereas prod uses t-degree
        f_j = sum_div_2(prod[2 * j])  # the coeff. of the jth term in <+|F(z)|+>
        f_mj = sum_div_2(prod[-2 * j])  # the coeff. of the -jth term in <+|F(z)|+>

        # Split the function <+|F(z)|+> into reciprocal A(Z) and anti-reciprocal B(Z) functions.
        recip = chop((f_j + f_mj) / 2)
        anti_recip = chop((f_j - f_mj) / mpc(0, 2))

        a[j] = recip
        a[-j] = recip
        b[j] = anti_recip
        b[-j] = -anti_recip

    # Update big_n in case the leading terms of a and b are zero
    big_n = max(j for j in range(big_n + 1) if a[j] != 0 or b[j] != 0)

    # Step 4: Convert a and b to FloatFunctions
    return InputFunction(a, Parity.EVEN, big_n), InputFunction(b, Parity.ODD, big_n)
